/*
 * Copyright (C) 2022-2025 United States Government as represented by the Administrator of the
 * National Aeronautics and Space Administration. All Rights Reserved.
 *
 * NASA acknowledges the SETI Institute's primary role in authoring and producing Ziggy, a Pipeline
 * Management System for Data Analysis Pipelines, under Cooperative Agreement Nos. NNX14AH97A,
 * 80NSSC18M0068 & 80NSSC21M0079.
 *
 * This file is available under the terms of the NASA Open Source Agreement (NOSA). You should have
 * received a copy of this agreement with the Ziggy source code; see the file LICENSE.pdf.
 *
 * Disclaimers
 *
 * No Warranty: THE SUBJECT SOFTWARE IS PROVIDED "AS IS" WITHOUT ANY WARRANTY OF ANY KIND, EITHER
 * EXPRESSED, IMPLIED, OR STATUTORY, INCLUDING, BUT NOT LIMITED TO, ANY WARRANTY THAT THE SUBJECT
 * SOFTWARE WILL CONFORM TO SPECIFICATIONS, ANY IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE, OR FREEDOM FROM INFRINGEMENT, ANY WARRANTY THAT THE SUBJECT SOFTWARE WILL BE
 * ERROR FREE, OR ANY WARRANTY THAT DOCUMENTATION, IF PROVIDED, WILL CONFORM TO THE SUBJECT
 * SOFTWARE. THIS AGREEMENT DOES NOT, IN ANY MANNER, CONSTITUTE AN ENDORSEMENT BY GOVERNMENT AGENCY
 * OR ANY PRIOR RECIPIENT OF ANY RESULTS, RESULTING DESIGNS, HARDWARE, SOFTWARE PRODUCTS OR ANY
 * OTHER APPLICATIONS RESULTING FROM USE OF THE SUBJECT SOFTWARE. FURTHER, GOVERNMENT AGENCY
 * DISCLAIMS ALL WARRANTIES AND LIABILITIES REGARDING THIRD-PARTY SOFTWARE, IF PRESENT IN THE
 * ORIGINAL SOFTWARE, AND DISTRIBUTES IT "AS IS."
 *
 * Waiver and Indemnity: RECIPIENT AGREES TO WAIVE ANY AND ALL CLAIMS AGAINST THE UNITED STATES
 * GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY PRIOR RECIPIENT. IF RECIPIENT'S
 * USE OF THE SUBJECT SOFTWARE RESULTS IN ANY LIABILITIES, DEMANDS, DAMAGES, EXPENSES OR LOSSES
 * ARISING FROM SUCH USE, INCLUDING ANY DAMAGES FROM PRODUCTS BASED ON, OR RESULTING FROM,
 * RECIPIENT'S USE OF THE SUBJECT SOFTWARE, RECIPIENT SHALL INDEMNIFY AND HOLD HARMLESS THE UNITED
 * STATES GOVERNMENT, ITS CONTRACTORS AND SUBCONTRACTORS, AS WELL AS ANY PRIOR RECIPIENT, TO THE
 * EXTENT PERMITTED BY LAW. RECIPIENT'S SOLE REMEDY FOR ANY SUCH MATTER SHALL BE THE IMMEDIATE,
 * UNILATERAL TERMINATION OF THIS AGREEMENT.
 */

package gov.nasa.ziggy.pipeline;

import java.util.List;

import org.apache.commons.configuration2.ImmutableConfiguration;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import gov.nasa.ziggy.pipeline.definition.Pipeline;
import gov.nasa.ziggy.pipeline.definition.PipelineNode;
import gov.nasa.ziggy.pipeline.definition.PipelineStepExecutor;
import gov.nasa.ziggy.pipeline.definition.PipelineTask;
import gov.nasa.ziggy.pipeline.definition.database.PipelineNodeOperations;
import gov.nasa.ziggy.pipeline.definition.database.PipelineOperations;
import gov.nasa.ziggy.pipeline.definition.database.PipelineTaskOperations;
import gov.nasa.ziggy.pipeline.step.AlgorithmPipelineStepExecutor;
import gov.nasa.ziggy.services.config.ZiggyConfiguration;
import gov.nasa.ziggy.util.AcceptableCatchBlock;
import gov.nasa.ziggy.util.AcceptableCatchBlock.Rationale;

public class PipelineTaskDebugger {

    private static final Logger log = LoggerFactory.getLogger(PipelineTaskDebugger.class);

    private static final String TASK_DEBUGGER_PREFIX = "ptd.";
    private static final String PIPELINE_TASK_ID_PROP = TASK_DEBUGGER_PREFIX + "pipelineTaskId";
    private static final String DEBUG_ASYNC_ENABLED_PROP = TASK_DEBUGGER_PREFIX
        + "debugAsyncEnabled";
    private static final String DEBUG_FROM_START_ENABLED_PROP = TASK_DEBUGGER_PREFIX
        + "debugFromStartEnabled";
    private static final String DEBUG_LAUNCH_PIPELINE_ENABLED_PROP = TASK_DEBUGGER_PREFIX
        + "debugLaunchPipelineEnabled";
    private static final String TASK_DIR_PROP = TASK_DEBUGGER_PREFIX + "taskDir";
    private static final String PERSISTABLE_OUTPUT_CLASS_NAME_PROP = TASK_DEBUGGER_PREFIX
        + "persistableClassName";
    private static final String PIPELINE_STEP_NAME_PROP = TASK_DEBUGGER_PREFIX + "pipelineStepName";
    private static final String PIPELINE_NAME_PROP = TASK_DEBUGGER_PREFIX + "pipelineName";

    private int pipelineTaskId;
    private boolean debugAsyncEnabled;
    private boolean debugFromStartEnabled;
    private boolean debugLaunchPipelineEnabled;
    private String taskDir;
    private String persistableOutputClassName;
    private String pipelineStepName;
    private String pipelineName;

    /*
     * Either set properties, ptd.*, or provide explicit values as the defaults in the config calls
     * in this constructor.
     */
    public PipelineTaskDebugger(ImmutableConfiguration config) {
        // used by all
        pipelineTaskId = config.getInt(PIPELINE_TASK_ID_PROP, 35);
        debugAsyncEnabled = config.getBoolean(DEBUG_ASYNC_ENABLED_PROP, true);
        debugFromStartEnabled = config.getBoolean(DEBUG_FROM_START_ENABLED_PROP, false);
        debugLaunchPipelineEnabled = config.getBoolean(DEBUG_LAUNCH_PIPELINE_ENABLED_PROP, false);

        if (isDebugAsyncEnabled()) {
            persistableOutputClassName = config.getString(PERSISTABLE_OUTPUT_CLASS_NAME_PROP,
                "gov.nasa.kepler.pa.PaOutputs");
            taskDir = config.getString(TASK_DIR_PROP);
        } else if (isDebugLaunchPipelineEnabled()) {
            pipelineStepName = config.getString(PIPELINE_STEP_NAME_PROP, "pa");
            pipelineName = config.getString(PIPELINE_NAME_PROP, "PHOTOMETRY_LC");
        }
    }

    public static void main(String[] args) {
        PipelineTaskDebugger pipelineTaskDebugger = new PipelineTaskDebugger(
            ZiggyConfiguration.getInstance());
        startProcessingThread(pipelineTaskDebugger);
        log.info("Completed");

        System.exit(0);
    }

    @AcceptableCatchBlock(rationale = Rationale.SYSTEM_EXIT)
    private static void startProcessingThread(final PipelineTaskDebugger pipelineTaskDebugger) {
        Thread thread = new Thread() {
            private PipelineTaskOperations pipelineTaskOperations = new PipelineTaskOperations();
            private PipelineNodeOperations pipelineNodeOperations = new PipelineNodeOperations();

            @Override
            @AcceptableCatchBlock(rationale = Rationale.SYSTEM_EXIT)
            public void run() {
                try {
                    PipelineTask task = pipelineTaskOperations
                        .pipelineTask(pipelineTaskDebugger.getPipelineTaskId());

                    dumpConfig(pipelineTaskDebugger, task);
                    if (pipelineTaskDebugger.isDebugAsyncEnabled()) {
                        debugAsyncLocalFromProcessOutputs(task);
                    } else if (pipelineTaskDebugger.isDebugFromStartEnabled()) {
                        debugFromStartOfTask(task);
                    } else {
                        debugDoTransition(task);
                    }
                    if (pipelineTaskDebugger.isDebugLaunchPipelineEnabled()) {
                        debugLaunchPipeline();
                    }
                } catch (Throwable e) {
                    processException(e);
                }
            }

            private void debugFromStartOfTask(PipelineTask pipelineTask) throws Exception {
                PipelineStepExecutor pipelineStepExecutor = pipelineTaskOperations
                    .pipelineStepExecutorImplementation(pipelineTask);
                pipelineStepExecutor.processTask();
            }

            private void debugAsyncLocalFromProcessOutputs(PipelineTask pipelineTask) {
                PipelineStepExecutor pipelineStepExecutor = pipelineTaskOperations
                    .pipelineStepExecutorImplementation(pipelineTask);
                Object pipelineStepName = pipelineStepExecutor.getPipelineStepName();
                if (!(pipelineStepName instanceof String)) {
                    throw new IllegalStateException(String.format(
                        "%s: invalid class, getPipelineStepName method did not return String",
                        pipelineTaskOperations.pipelineStep(pipelineTask)
                            .getPipelineStepExecutorClass()
                            .getClassName()));
                }

                if (!(pipelineStepExecutor instanceof AlgorithmPipelineStepExecutor)) {
                    log.error(
                        "Configured pipelineStepExecutor must implement AlgorithmPipelineStepExecutor");
                }
            }

            private void debugDoTransition(PipelineTask pipelineTask) {
                PipelineExecutor pipelineExecutor = new PipelineExecutor();
                pipelineExecutor.transitionToNextInstanceNode(
                    pipelineTaskOperations.pipelineInstanceNode(pipelineTask));
            }

            private void debugLaunchPipeline() {
                PipelineOperations pipelineOperations = new PipelineOperations();
                Pipeline pipeline = pipelineOperations
                    .pipeline(pipelineTaskDebugger.getPipelineName());

                PipelineNode pipelineNode = getPipelineNode(pipelineOperations.rootNodes(pipeline));

                new PipelineExecutor().launch(pipeline, "instanceName", pipelineNode, pipelineNode,
                    null);
            }

            private PipelineNode getPipelineNode(List<PipelineNode> nodes) {
                for (PipelineNode node : nodes) {
                    if (node.getPipelineStepName()
                        .equals(pipelineTaskDebugger.getPipelineStepName())) {
                        return node;
                    }

                    PipelineNode nodeFromRecursiveCall = getPipelineNode(
                        pipelineNodeOperations.nextNodes(node));
                    if (nodeFromRecursiveCall != null) {
                        return nodeFromRecursiveCall;
                    }
                }
                return null;
            }
        };
        thread.start();
        try {
            thread.join();
        } catch (InterruptedException e) {
            System.exit(2);
        }
    }

    private static void processException(Throwable e) {
        log.error("Caught exception: ", e);
        e.printStackTrace();
        log.error("Terminated");
        System.exit(1);
    }

    private static void dumpConfig(PipelineTaskDebugger taskDebugger, PipelineTask pipelineTask) {

        log.info(String.format("%s=%s\n", PIPELINE_TASK_ID_PROP, taskDebugger.getPipelineTaskId()));
        log.info(
            String.format("%s=%s\n", DEBUG_ASYNC_ENABLED_PROP, taskDebugger.isDebugAsyncEnabled()));
        log.info(String.format("%s=%s\n", DEBUG_FROM_START_ENABLED_PROP,
            taskDebugger.isDebugFromStartEnabled()));
        log.info(String.format("%s=%s\n", DEBUG_LAUNCH_PIPELINE_ENABLED_PROP,
            taskDebugger.isDebugLaunchPipelineEnabled()));
        log.info(String.format("%s=%s\n", TASK_DIR_PROP, taskDebugger.getTaskDir()));
        log.info(String.format("%s=%s\n", PERSISTABLE_OUTPUT_CLASS_NAME_PROP,
            taskDebugger.getPersistableOutputClassName()));
        log.info(
            String.format("%s=%s\n", PIPELINE_STEP_NAME_PROP, taskDebugger.getPipelineStepName()));
        log.info(String.format("%s=%s\n", PIPELINE_NAME_PROP, taskDebugger.getPipelineName()));

        log.info(String.format("%s=%s\n", ZiggyConfiguration.PIPELINE_CONFIG_PATH_ENV,
            System.getenv(ZiggyConfiguration.PIPELINE_CONFIG_PATH_ENV)));

        PipelineStepExecutor pipelineStepExecutor = new PipelineTaskOperations()
            .pipelineStepExecutorImplementation(pipelineTask);
        Object pipelineStepName = pipelineStepExecutor.getPipelineStepName();
        if (pipelineStepName instanceof String) {
            log.info(String.format("PIPELINE_STEP_NAME=%s\n", (String) pipelineStepName));
        }
    }

    public int getPipelineTaskId() {
        return pipelineTaskId;
    }

    public boolean isDebugAsyncEnabled() {
        return debugAsyncEnabled;
    }

    public boolean isDebugFromStartEnabled() {
        return debugFromStartEnabled;
    }

    public boolean isDebugLaunchPipelineEnabled() {
        return debugLaunchPipelineEnabled;
    }

    public String getTaskDir() {
        return taskDir;
    }

    public String getPersistableOutputClassName() {
        return persistableOutputClassName;
    }

    public String getPipelineProperties() {
        return ZiggyConfiguration.PIPELINE_CONFIG_PATH_ENV;
    }

    public String getPipelineStepName() {
        return pipelineStepName;
    }

    public String getPipelineName() {
        return pipelineName;
    }
}
