package gov.nasa.ziggy.pipeline.step;

import java.io.File;
import java.nio.file.Path;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import gov.nasa.ziggy.pipeline.definition.PipelineTask;
import gov.nasa.ziggy.pipeline.definition.database.ParametersOperations;
import gov.nasa.ziggy.pipeline.definition.database.PipelineTaskOperations;
import gov.nasa.ziggy.pipeline.step.remote.RemoteAlgorithmExecutor;
import gov.nasa.ziggy.util.AcceptableCatchBlock;
import gov.nasa.ziggy.util.AcceptableCatchBlock.Rationale;

/**
 * This class manages the lifecycle of a pipeline algorithm.
 *
 * @author Todd Klaus
 * @author PT
 */
public class AlgorithmLifecycleManager implements AlgorithmLifecycle {
    private static final Logger log = LoggerFactory.getLogger(AlgorithmLifecycleManager.class);

    private TaskDirectoryManager taskDirManager;
    private Path taskDir;

    private PipelineTask pipelineTask;
    private AlgorithmExecutor executor;
    private PipelineTaskOperations pipelineTaskOperations = new PipelineTaskOperations();
    private ParametersOperations parametersOperations = new ParametersOperations();

    public AlgorithmLifecycleManager(PipelineTask pipelineTask) {
        this.pipelineTask = pipelineTask;
        taskDirManager = new TaskDirectoryManager(pipelineTask);

        // We need an executor at construction time, though it may get replaced later.
        executor = AlgorithmExecutor.newInstance(pipelineTask);
    }

    @Override
    public void executeAlgorithm(TaskConfiguration inputs) {

        // Replace the pipeline task and the executor now, since we have new information
        // about the task's subtask counts.
        executor = AlgorithmExecutor.newInstance(pipelineTask);
        executor.submitAlgorithm(inputs);
    }

    @Override
    public void doPostProcessing() {
    }

    @Override
    @AcceptableCatchBlock(rationale = Rationale.EXCEPTION_CHAIN)
    public File getTaskDir(boolean cleanExisting) {
        return allocateTaskDir(cleanExisting);
    }

    @Override
    public boolean isRemote() {
        return executor instanceof RemoteAlgorithmExecutor;
    }

    @Override
    public AlgorithmExecutor getExecutor() {
        return executor;
    }

    /**
     * Allocate the working directory using the specified prefix.
     *
     * @param workingDirNamePrefix
     * @param pipelineTask
     * @return
     */
    private File allocateTaskDir(boolean cleanExisting) {

        if (taskDirManager == null) {
            taskDirManager = new TaskDirectoryManager(pipelineTask);
        }

        if (taskDir == null) {
            taskDir = taskDirManager.allocateTaskDir(cleanExisting);
            log.info("defaultWorkingDir={}", taskDir);
        }
        return taskDir.toFile();
    }

    PipelineTaskOperations pipelineTaskOperations() {
        return pipelineTaskOperations;
    }

    ParametersOperations parametersOperations() {
        return parametersOperations;
    }
}
